/*
    dsproxy_main.c - OSS trampoline device driver

    Copyright (C) 2000 Ram Sivrasubramanium <ram@linux-workshop.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <linux/string.h>
#include <linux/time.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/sched.h>
#include <asm/segment.h>
#include <linux/poll.h>
#include <asm/poll.h>
#include <linux/version.h>
#include <asm/uaccess.h>

#include "dsproxy.h"

MODULE_AUTHOR("Ram Sivrasubramanium <ram@linux-workshop.com>");
MODULE_DESCRIPTION("OSS Trampoline");


struct dsproxy_s *emhead=NULL;
struct semaphore global_sem;

static int dsproxy_io_open(struct inode* inode, struct file* filp) 
{
int rc;
int subdevice=MINOR(inode->i_rdev);

switch(subdevice) 
	{
    	case DSPROXY_SUBDEVICE_AUDIO:
		DEBUG(printk("dsproxy_main.o: open() DSPROXY_SUBDEVICE_AUDIO \n"));
		rc =  dsproxy_audio_open(inode, filp);
		break;

	case DSPROXY_SUBDEVICE_MIXER:
		DEBUG(printk("dsproxy_main.o: open() DSPROXY_SUBDEVICE_MIXER \n"));
		rc =  dsproxy_mixer_open(inode, filp);
		break;

    	default:
		DEBUG(printk("dsproxy_main.o: open() *unknown* \n"));
		return(-ENODEV);
		break;
    	}

if(!rc)
	{
	MOD_INC_USE_COUNT;
	}
 
return(rc);
}


int dsproxy_io_release(struct inode* inode, struct file* filp) 
{
int rc;
int subdevice = MINOR(inode->i_rdev);
struct dsproxy_s *em = filp->private_data;
    
switch(subdevice) 
	{
    	case DSPROXY_SUBDEVICE_AUDIO:
		rc = dsproxy_audio_release(em, filp, subdevice);
		break;

	case DSPROXY_SUBDEVICE_MIXER:
		rc = dsproxy_mixer_release(em, filp, subdevice);
		break;

    	default:
		return(-EINVAL);
    	}

if(!rc)
	{
	MOD_DEC_USE_COUNT;
	}
	
return(rc);
}
   

static int dsproxy_io_read(struct file *file, char * buf, size_t count, loff_t *ppos)
{
struct dsproxy_s *em = file->private_data;
int subdevice = MINOR(file->f_dentry->d_inode->i_rdev);
int rc;
    
DEBUG(printk("dsproxy_main.o: dsproxy_io_read() - in\n"));

switch (subdevice) 
	{
    	case DSPROXY_SUBDEVICE_AUDIO:
		if(em->r_inuse)
			{
			down(&em->forkread);		/* make forked reads sequential.  a user should *never* do this, but this is to
							   keep things internally sane just in case.. */
			rc=dsproxy_audio_read(file, em,buf,count,ppos);
			up(&em->forkread);
			DEBUG(printk("dsproxy_main.o: dsproxy_io_read() - out ok\n"));
			return(rc);
			}
			else
			{
			DEBUG(printk("dsproxy_main.o: dsproxy_io_read() - out -EPIPE\n"));
			return(-EPIPE);
			}

	case DSPROXY_SUBDEVICE_MIXER:
    	default:
			DEBUG(printk("dsproxy_main.o: dsproxy_io_read() - out -EPERM\n"));
		return(-EPERM);
    	}
}



static int dsproxy_io_write(struct file *file, const char * buf, size_t count, loff_t *ppos)
{
struct dsproxy_s *em = file->private_data;
int subdevice = MINOR(file->f_dentry->d_inode->i_rdev);
int rc;
    
DEBUG(printk("dsproxy_main.o: dsproxy_io_write() \n"));

switch (subdevice) 
	{
    	case DSPROXY_SUBDEVICE_AUDIO:
		if(em->r_inuse)
			{
			down(&em->forkwrite);		/* make forked writes sequential to keep buffer scrambling from occurring */
			rc=dsproxy_audio_write(file, em,buf,count,ppos);
			up(&em->forkwrite);
			return(rc);
			}
			else
			{
			return(-EPIPE);
			}

	case DSPROXY_SUBDEVICE_MIXER:
    	default:
		return(-EPERM);
    	}
}


static int dsproxy_io_ioctl(struct inode* inode, struct file* filp, unsigned int cmd, unsigned long arg)
{
struct dsproxy_s *em = filp->private_data;
int subdevice = MINOR(inode->i_rdev);
int rc;

DEBUG(printk("dsproxy_main.o: dsproxy_io_ioctl() \n"));

switch(subdevice) 
	{
    	case DSPROXY_SUBDEVICE_AUDIO:
		if(em->r_inuse)
			{
			if((filp->f_flags&O_ACCMODE) == O_WRONLY)
				{
				down(&em->forkwrite);
				rc=dsproxy_audio_ioctl(filp,em,cmd,arg);
				up(&em->forkwrite);
				return(rc);
				}
				else
				{
				return(-EPERM);
				}
			}
			else
			{
			return(-EPIPE);
			}

	case DSPROXY_SUBDEVICE_MIXER:
		return(dsproxy_mixer_ioctl(filp,em,cmd,arg));

	default:
		return(-ENODEV);
	}
}


unsigned int dsproxy_io_poll(struct file *filp, poll_table *wait)
{
struct dsproxy_s *em = filp->private_data;

switch(em->subdevice) 
	{
    	case DSPROXY_SUBDEVICE_AUDIO:
		if(em->r_inuse)
			{
			return(dsproxy_audio_poll(filp,wait));
			}
			else
			{
			return(-EPIPE);
			}
	
	case DSPROXY_SUBDEVICE_MIXER:
    	default:
		return(-EPERM);
    	}
}


loff_t dsproxy_io_llseek(struct file *filp, loff_t offset, int whence)
{
return(-ESPIPE);
}


/*
 * tables/linking into the linux kernel..
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
static struct file_operations dsproxy_fops = {
  dsproxy_io_llseek,
  dsproxy_io_read,
  dsproxy_io_write,
  NULL,
  dsproxy_io_poll,
  dsproxy_io_ioctl,
  NULL,
  dsproxy_io_open,
  NULL,
  dsproxy_io_release,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL };
#else
static struct file_operations dsproxy_fops = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
	THIS_MODULE,
#endif
  dsproxy_io_llseek,
  dsproxy_io_read,
  dsproxy_io_write,
  NULL,
  dsproxy_io_poll,
  dsproxy_io_ioctl,
  NULL,
  dsproxy_io_open,
  NULL,
  dsproxy_io_release,
  NULL,
  NULL,
  NULL };
#endif


int init_module(void)
{
memset(&global_sem, 0, sizeof(struct semaphore));
sema_init(&global_sem, 1);

if(register_chrdev(DSPROXY_MAJOR, DSPROXY_LOGNAME, &dsproxy_fops)) 
	{
        printk(KERN_ERR "dsproxy_main.o: unable to get major %d\n", DSPROXY_MAJOR);
        return(-EBUSY);
    	}

printk("dsproxy_main.o: installed module\n");
return(0);
}


void cleanup_module(void) 
{
printk("dsproxy_main.o: removed module\n");
unregister_chrdev(DSPROXY_MAJOR,DSPROXY_LOGNAME);
}
